@@ -60,10 +60,6 @@
 
 #include "../sspi.h"
 #include "../../log.h"
-#define TAG WINPR_TAG("sspi.Kerberos")
-
-#define KRB_TGT_REQ 16
-#define KRB_TGT_REP 17
 
 const SecPkgInfoA KERBEROS_SecPkgInfoA = {
 	0x000F3BBF,                 /* fCapabilities */
@@ -87,6 +83,10 @@ const SecPkgInfoW KERBEROS_SecPkgInfoW = {
 };
 
 #ifdef WITH_KRB5
+#define TAG WINPR_TAG("sspi.Kerberos")
+
+#define KRB_TGT_REQ 16
+#define KRB_TGT_REP 17
 
 enum KERBEROS_STATE
 {
@@ -127,9 +127,6 @@ static const WinPrAsn1_OID kerberos_OID = { 9, (void*)"\x2a\x86\x48\x86\xf7\x12\
 static const WinPrAsn1_OID kerberos_u2u_OID = { 10,
 	                                            (void*)"\x2a\x86\x48\x86\xf7\x12\x01\x02\x02\x03" };
 
-#define krb_log_exec_bool(fkt, ctx, ...)                                                     \
-	kerberos_log_msg(ctx, fkt(ctx, ##__VA_ARGS__) ? KRB5KRB_ERR_GENERIC : 0, #fkt, __FILE__, \
-	                 __func__, __LINE__)
 #define krb_log_exec(fkt, ctx, ...) \
 	kerberos_log_msg(ctx, fkt(ctx, ##__VA_ARGS__), #fkt, __FILE__, __func__, __LINE__)
 #define krb_log_exec_ptr(fkt, ctx, ...) \
@@ -150,8 +147,7 @@ static krb5_error_code kerberos_log_msg(krb5_context ctx, krb5_error_code code,
 			if (WLog_IsLevelActive(log, level))
 			{
 				const char* msg = krb5_get_error_message(ctx, code);
-				WLog_PrintMessage(log, WLOG_MESSAGE_TEXT, level, line, file, fkt, "%s (%s [%d])",
-				                  what, msg, code);
+				WLog_PrintTextMessage(log, level, line, file, fkt, "%s (%s [%d])", what, msg, code);
 				krb5_free_error_message(ctx, msg);
 			}
 		}
@@ -222,7 +218,7 @@ static krb5_error_code krb5_prompter(krb5_context context, void* data,
 	return 0;
 }
 
-static INLINE krb5glue_key get_key(struct krb5glue_keyset* keyset)
+static inline krb5glue_key get_key(struct krb5glue_keyset* keyset)
 {
 	return keyset->acceptor_key    ? keyset->acceptor_key
 	       : keyset->initiator_key ? keyset->initiator_key
@@ -248,20 +244,52 @@ static BOOL isValidIP(const char* ipAddress)
 	return isValidIPv4(ipAddress) || isValidIPv6(ipAddress);
 }
 
-static int build_krbtgt(krb5_context ctx, krb5_data* realm, krb5_principal* ptarget)
+#if defined(WITH_KRB5_MIT)
+WINPR_ATTR_MALLOC(free, 1)
+WINPR_ATTR_NODISCARD
+static char* get_realm_name(krb5_data realm, size_t* plen)
+{
+	WINPR_ASSERT(plen);
+	*plen = 0;
+	if ((realm.length <= 0) || (!realm.data))
+		return NULL;
+
+	char* name = NULL;
+	(void)winpr_asprintf(&name, plen, "krbtgt/%*s@%*s", realm.length, realm.data, realm.length,
+	                     realm.data);
+	return name;
+}
+#elif defined(WITH_KRB5_HEIMDAL)
+WINPR_ATTR_MALLOC(free, 1)
+WINPR_ATTR_NODISCARD
+static char* get_realm_name(Realm realm, size_t* plen)
+{
+	WINPR_ASSERT(plen);
+	*plen = 0;
+	if (!realm)
+		return NULL;
+
+	char* name = NULL;
+	(void)winpr_asprintf(&name, plen, "krbtgt/%s@%s", realm, realm);
+	return name;
+}
+#endif
+
+static int build_krbtgt(krb5_context ctx, krb5_principal principal, krb5_principal* ptarget)
 {
 	/* "krbtgt/" + realm + "@" + realm */
 	size_t len = 0;
-	char* name = NULL;
 	krb5_error_code rv = KRB5_CC_NOMEM;
 
-	(void)winpr_asprintf(&name, &len, "krbtgt/%s@%s", realm->data, realm->data);
+	char* name = get_realm_name(principal->realm, &len);
 	if (!name || (len == 0))
 		goto fail;
 
-	krb5_principal target = { 0 };
-	rv = krb5_parse_name(ctx, name, &target);
-	*ptarget = target;
+	{
+		krb5_principal target = { 0 };
+		rv = krb5_parse_name(ctx, name, &target);
+		*ptarget = target;
+	}
 fail:
 	free(name);
 	return rv;
@@ -270,9 +298,10 @@ fail:
 #endif /* WITH_KRB5 */
 
 static SECURITY_STATUS SEC_ENTRY kerberos_AcquireCredentialsHandleA(
-    SEC_CHAR* pszPrincipal, WINPR_ATTR_UNUSED SEC_CHAR* pszPackage, ULONG fCredentialUse,
-    WINPR_ATTR_UNUSED void* pvLogonID, void* pAuthData, WINPR_ATTR_UNUSED SEC_GET_KEY_FN pGetKeyFn,
-    WINPR_ATTR_UNUSED void* pvGetKeyArgument, PCredHandle phCredential,
+    WINPR_ATTR_UNUSED SEC_CHAR* pszPrincipal, WINPR_ATTR_UNUSED SEC_CHAR* pszPackage,
+    WINPR_ATTR_UNUSED ULONG fCredentialUse, WINPR_ATTR_UNUSED void* pvLogonID,
+    WINPR_ATTR_UNUSED void* pAuthData, WINPR_ATTR_UNUSED SEC_GET_KEY_FN pGetKeyFn,
+    WINPR_ATTR_UNUSED void* pvGetKeyArgument, WINPR_ATTR_UNUSED PCredHandle phCredential,
     WINPR_ATTR_UNUSED PTimeStamp ptsExpiry)
 {
 #ifdef WITH_KRB5
@@ -340,6 +369,7 @@ static SECURITY_STATUS SEC_ENTRY kerberos_AcquireCredentialsHandleA(
 
 		if (rv)
 			goto cleanup;
+		WINPR_ASSERT(principal);
 	}
 
 	if (krb_settings && krb_settings->cache)
@@ -370,7 +400,12 @@ static SECURITY_STATUS SEC_ENTRY kerberos_AcquireCredentialsHandleA(
 				goto cleanup;
 		}
 		else
+		{
+			if (krb_log_exec(krb5_cc_default, ctx, &ccache))
+				goto cleanup;
 			own_ccache = FALSE;
+		}
+		WINPR_ASSERT(ccache);
 	}
 	else if (fCredentialUse & SECPKG_CRED_OUTBOUND)
 	{
@@ -379,6 +414,7 @@ static SECURITY_STATUS SEC_ENTRY kerberos_AcquireCredentialsHandleA(
 			goto cleanup;
 		if (krb_log_exec(krb5_cc_get_principal, ctx, ccache, &principal))
 			goto cleanup;
+		WINPR_ASSERT(ccache);
 		own_ccache = FALSE;
 	}
 	else
@@ -393,6 +429,7 @@ static SECURITY_STATUS SEC_ENTRY kerberos_AcquireCredentialsHandleA(
 			if (krb_log_exec(krb5_cc_resolve, ctx, krb_settings->cache, &ccache))
 				goto cleanup;
 		}
+		WINPR_ASSERT(ccache);
 	}
 
 	if (krb_settings && krb_settings->keytab)
@@ -418,7 +455,10 @@ static SECURITY_STATUS SEC_ENTRY kerberos_AcquireCredentialsHandleA(
 		matchCreds.times.endtime += 60;
 		matchCreds.client = principal;
 
-		if (krb_log_exec(build_krbtgt, ctx, &principal->realm, &matchCreds.server))
+		WINPR_ASSERT(principal);
+		WINPR_ASSERT(ctx);
+		WINPR_ASSERT(ccache);
+		if (krb_log_exec(build_krbtgt, ctx, principal, &matchCreds.server))
 			goto cleanup;
 
 		int rv = krb5_cc_retrieve_cred(ctx, ccache, matchFlags, &matchCreds, &creds);
@@ -539,7 +579,8 @@ static void credentials_unref(KRB_CREDENTIALS* credentials)
 }
 #endif
 
-static SECURITY_STATUS SEC_ENTRY kerberos_FreeCredentialsHandle(PCredHandle phCredential)
+static SECURITY_STATUS
+    SEC_ENTRY kerberos_FreeCredentialsHandle(WINPR_ATTR_UNUSED PCredHandle phCredential)
 {
 #ifdef WITH_KRB5
 	KRB_CREDENTIALS* credentials = sspi_SecureHandleGetLowerPointer(phCredential);
@@ -556,7 +597,8 @@ static SECURITY_STATUS SEC_ENTRY kerberos_FreeCredentialsHandle(PCredHandle phCr
 }
 
 static SECURITY_STATUS SEC_ENTRY kerberos_QueryCredentialsAttributesW(
-    WINPR_ATTR_UNUSED PCredHandle phCredential, ULONG ulAttribute, WINPR_ATTR_UNUSED void* pBuffer)
+    WINPR_ATTR_UNUSED PCredHandle phCredential, WINPR_ATTR_UNUSED ULONG ulAttribute,
+    WINPR_ATTR_UNUSED void* pBuffer)
 {
 #ifdef WITH_KRB5
 	switch (ulAttribute)
@@ -675,7 +717,7 @@ static BOOL append(char* dst, size_t dstSize, const char* src)
 	const size_t slen = strlen(src);
 	if (dlen + slen >= dstSize)
 		return FALSE;
-	if (!strncat(dst, src, slen))
+	if (!strncat(dst, src, dstSize - dlen))
 		return FALSE;
 	return TRUE;
 }
@@ -683,48 +725,54 @@ static BOOL append(char* dst, size_t dstSize, const char* src)
 static BOOL kerberos_rd_tgt_req_tag2(WinPrAsn1Decoder* dec, char* buf, size_t len)
 {
 	BOOL rc = FALSE;
-	WinPrAsn1Decoder seq = { 0 };
+	WinPrAsn1Decoder seq = { .encoding = WINPR_ASN1_BER, { 0 } };
 
 	/* server-name [2] PrincipalName (SEQUENCE) */
 	if (!WinPrAsn1DecReadSequence(dec, &seq))
 		goto end;
 
 	/* name-type [0] INTEGER */
-	BOOL error = FALSE;
-	WinPrAsn1_INTEGER val = 0;
-	if (!WinPrAsn1DecReadContextualInteger(&seq, 0, &error, &val))
-		goto end;
-
-	/* name-string [1] SEQUENCE OF GeneralString */
-	if (!WinPrAsn1DecReadContextualSequence(&seq, 1, &error, dec))
-		goto end;
-
-	WinPrAsn1_tag tag = 0;
-	BOOL first = TRUE;
-	while (WinPrAsn1DecPeekTag(dec, &tag))
 	{
-		BOOL success = FALSE;
-		char* lstr = NULL;
-		if (!WinPrAsn1DecReadGeneralString(dec, &lstr))
-			goto fail;
-
-		if (!first)
+		BOOL error = FALSE;
 		{
-			if (!append(buf, len, "/"))
-				goto fail;
+			WinPrAsn1_INTEGER val = 0;
+			if (!WinPrAsn1DecReadContextualInteger(&seq, 0, &error, &val))
+				goto end;
 		}
-		first = FALSE;
 
-		if (!append(buf, len, lstr))
-			goto fail;
-
-		success = TRUE;
-	fail:
-		free(lstr);
-		if (!success)
+		/* name-string [1] SEQUENCE OF GeneralString */
+		if (!WinPrAsn1DecReadContextualSequence(&seq, 1, &error, dec))
 			goto end;
 	}
 
+	{
+		WinPrAsn1_tag tag = 0;
+		BOOL first = TRUE;
+		while (WinPrAsn1DecPeekTag(dec, &tag))
+		{
+			BOOL success = FALSE;
+			char* lstr = NULL;
+			if (!WinPrAsn1DecReadGeneralString(dec, &lstr))
+				goto fail;
+
+			if (!first)
+			{
+				if (!append(buf, len, "/"))
+					goto fail;
+			}
+			first = FALSE;
+
+			if (!append(buf, len, lstr))
+				goto fail;
+
+			success = TRUE;
+		fail:
+			free(lstr);
+			if (!success)
+				goto end;
+		}
+	}
+
 	rc = TRUE;
 end:
 	return rc;
@@ -762,7 +810,7 @@ static BOOL kerberos_rd_tgt_req(WinPrAsn1Decoder* dec, char** target)
 	if (len == 0)
 		return TRUE;
 
-	WinPrAsn1Decoder dec2 = { 0 };
+	WinPrAsn1Decoder dec2 = { .encoding = WINPR_ASN1_BER, { 0 } };
 	WinPrAsn1_tagId tag = 0;
 	if (WinPrAsn1DecReadContextualTag(dec, &tag, &dec2) == 0)
 		return FALSE;
@@ -808,7 +856,7 @@ static BOOL kerberos_rd_tgt_rep(WinPrAsn1Decoder* dec, krb5_data* ticket)
 		return FALSE;
 
 	/* ticket [2] Ticket */
-	WinPrAsn1Decoder asnTicket = { 0 };
+	WinPrAsn1Decoder asnTicket = { .encoding = WINPR_ASN1_BER, { 0 } };
 	WinPrAsn1_tagId tag = 0;
 	if (WinPrAsn1DecReadContextualTag(dec, &tag, &asnTicket) == 0)
 		return FALSE;
@@ -836,11 +884,11 @@ static BOOL kerberos_rd_tgt_token(const sspi_gss_data* token, char** target, krb
 	if (target)
 		*target = NULL;
 
-	WinPrAsn1Decoder der = { 0 };
+	WinPrAsn1Decoder der = { .encoding = WINPR_ASN1_BER, { 0 } };
 	WinPrAsn1Decoder_InitMem(&der, WINPR_ASN1_DER, (BYTE*)token->data, token->length);
 
 	/* KERB-TGT-REQUEST (SEQUENCE) */
-	WinPrAsn1Decoder seq = { 0 };
+	WinPrAsn1Decoder seq = { .encoding = WINPR_ASN1_BER, { 0 } };
 	if (!WinPrAsn1DecReadSequence(&der, &seq))
 		return FALSE;
 
@@ -864,8 +912,6 @@ static BOOL kerberos_rd_tgt_token(const sspi_gss_data* token, char** target, krb
 	return FALSE;
 }
 
-#endif /* WITH_KRB5 */
-
 static BOOL kerberos_hash_channel_bindings(WINPR_DIGEST_CTX* md5, SEC_CHANNEL_BINDINGS* bindings)
 {
 	BYTE buf[4];
@@ -908,10 +954,14 @@ static BOOL kerberos_hash_channel_bindings(WINPR_DIGEST_CTX* md5, SEC_CHANNEL_BI
 	return TRUE;
 }
 
+#endif /* WITH_KRB5 */
+
 static SECURITY_STATUS SEC_ENTRY kerberos_InitializeSecurityContextA(
-    PCredHandle phCredential, PCtxtHandle phContext, SEC_CHAR* pszTargetName, ULONG fContextReq,
-    WINPR_ATTR_UNUSED ULONG Reserved1, WINPR_ATTR_UNUSED ULONG TargetDataRep, PSecBufferDesc pInput,
-    WINPR_ATTR_UNUSED ULONG Reserved2, PCtxtHandle phNewContext, PSecBufferDesc pOutput,
+    WINPR_ATTR_UNUSED PCredHandle phCredential, WINPR_ATTR_UNUSED PCtxtHandle phContext,
+    WINPR_ATTR_UNUSED SEC_CHAR* pszTargetName, WINPR_ATTR_UNUSED ULONG fContextReq,
+    WINPR_ATTR_UNUSED ULONG Reserved1, WINPR_ATTR_UNUSED ULONG TargetDataRep,
+    WINPR_ATTR_UNUSED PSecBufferDesc pInput, WINPR_ATTR_UNUSED ULONG Reserved2,
+    WINPR_ATTR_UNUSED PCtxtHandle phNewContext, WINPR_ATTR_UNUSED PSecBufferDesc pOutput,
     WINPR_ATTR_UNUSED ULONG* pfContextAttr, WINPR_ATTR_UNUSED PTimeStamp ptsExpiry)
 {
 #ifdef WITH_KRB5
@@ -1254,10 +1304,127 @@ static SECURITY_STATUS SEC_ENTRY kerberos_InitializeSecurityContextW(
 	return status;
 }
 
+#ifdef WITH_KRB5
+static BOOL retrieveTgtForPrincipal(KRB_CREDENTIALS* credentials, krb5_principal principal,
+                                    krb5_creds* creds)
+{
+	BOOL ret = FALSE;
+	krb5_kt_cursor cur = { 0 };
+	krb5_keytab_entry entry = { 0 };
+	if (krb_log_exec(krb5_kt_start_seq_get, credentials->ctx, credentials->keytab, &cur))
+		goto cleanup;
+
+	do
+	{
+		krb5_error_code rv =
+		    krb_log_exec(krb5_kt_next_entry, credentials->ctx, credentials->keytab, &entry, &cur);
+		if (rv == KRB5_KT_END)
+			break;
+		if (rv != 0)
+			goto cleanup;
+
+		if (krb5_principal_compare(credentials->ctx, principal, entry.principal))
+			break;
+		rv = krb_log_exec(krb5glue_free_keytab_entry_contents, credentials->ctx, &entry);
+		memset(&entry, 0, sizeof(entry));
+		if (rv)
+			goto cleanup;
+	} while (1);
+
+	if (krb_log_exec(krb5_kt_end_seq_get, credentials->ctx, credentials->keytab, &cur))
+		goto cleanup;
+
+	if (!entry.principal)
+		goto cleanup;
+
+	/* Get the TGT */
+	if (krb_log_exec(krb5_get_init_creds_keytab, credentials->ctx, creds, entry.principal,
+	                 credentials->keytab, 0, NULL, NULL))
+		goto cleanup;
+
+	ret = TRUE;
+
+cleanup:
+	return ret;
+}
+
+static BOOL retrieveSomeTgt(KRB_CREDENTIALS* credentials, const char* target, krb5_creds* creds)
+{
+	BOOL ret = TRUE;
+	krb5_principal target_princ = { 0 };
+	char* default_realm = NULL;
+
+	krb5_error_code rv =
+	    krb_log_exec(krb5_parse_name_flags, credentials->ctx, target, 0, &target_princ);
+	if (rv)
+		return FALSE;
+
+#if defined(WITH_KRB5_HEIMDAL)
+	if (!target_princ->realm)
+	{
+		rv = krb_log_exec(krb5_get_default_realm, credentials->ctx, &default_realm);
+		if (rv)
+			goto out;
+
+		target_princ->realm = default_realm;
+	}
+#else
+	if (!target_princ->realm.length)
+	{
+		rv = krb_log_exec(krb5_get_default_realm, credentials->ctx, &default_realm);
+		if (rv)
+			goto out;
+
+		target_princ->realm.data = default_realm;
+		target_princ->realm.length = (unsigned int)strlen(default_realm);
+	}
+#endif
+
+	/*
+	 * First try with the account service. We were requested with something like
+	 * TERMSRV/<host>@<realm>, let's see if we have that in our keytab and if we're able
+	 * to retrieve a TGT with that entry
+	 *
+	 */
+	if (retrieveTgtForPrincipal(credentials, target_princ, creds))
+		goto out;
+
+	ret = FALSE;
+
+#if defined(WITH_KRB5_MIT)
+	/*
+	 * if it's not working let's try with <host>$@<REALM> (note the dollar)
+	 */
+	{
+		char hostDollar[300] = { 0 };
+		if (target_princ->length < 2)
+			goto out;
+
+		(void)snprintf(hostDollar, sizeof(hostDollar) - 1, "%s$@%s", target_princ->data[1].data,
+		               target_princ->realm.data);
+		krb5_free_principal(credentials->ctx, target_princ);
+
+		rv = krb_log_exec(krb5_parse_name_flags, credentials->ctx, hostDollar, 0, &target_princ);
+		if (rv)
+			return FALSE;
+	}
+	ret = retrieveTgtForPrincipal(credentials, target_princ, creds);
+#endif
+
+out:
+	if (default_realm)
+		krb5_free_default_realm(credentials->ctx, default_realm);
+
+	krb5_free_principal(credentials->ctx, target_princ);
+	return ret;
+}
+#endif
+
 static SECURITY_STATUS SEC_ENTRY kerberos_AcceptSecurityContext(
-    PCredHandle phCredential, PCtxtHandle phContext, PSecBufferDesc pInput,
-    WINPR_ATTR_UNUSED ULONG fContextReq, WINPR_ATTR_UNUSED ULONG TargetDataRep,
-    PCtxtHandle phNewContext, PSecBufferDesc pOutput, ULONG* pfContextAttr,
+    WINPR_ATTR_UNUSED PCredHandle phCredential, WINPR_ATTR_UNUSED PCtxtHandle phContext,
+    WINPR_ATTR_UNUSED PSecBufferDesc pInput, WINPR_ATTR_UNUSED ULONG fContextReq,
+    WINPR_ATTR_UNUSED ULONG TargetDataRep, WINPR_ATTR_UNUSED PCtxtHandle phNewContext,
+    WINPR_ATTR_UNUSED PSecBufferDesc pOutput, WINPR_ATTR_UNUSED ULONG* pfContextAttr,
     WINPR_ATTR_UNUSED PTimeStamp ptsExpity)
 {
 #ifdef WITH_KRB5
@@ -1272,11 +1439,7 @@ static SECURITY_STATUS SEC_ENTRY kerberos_AcceptSecurityContext(
 	krb5_flags ap_flags = 0;
 	krb5glue_authenticator authenticator = NULL;
 	char* target = NULL;
-	char* sname = NULL;
-	char* realm = NULL;
-	krb5_kt_cursor cur = { 0 };
 	krb5_keytab_entry entry = { 0 };
-	krb5_principal principal = NULL;
 	krb5_creds creds = { 0 };
 
 	/* behave like windows SSPIs that don't want empty context */
@@ -1325,58 +1488,7 @@ static SECURITY_STATUS SEC_ENTRY kerberos_AcceptSecurityContext(
 		if (!kerberos_rd_tgt_token(&input_token, &target, NULL))
 			goto bad_token;
 
-		if (target)
-		{
-			if (*target != 0 && *target != '@')
-				sname = target;
-			realm = strchr(target, '@');
-			if (realm)
-				realm++;
-		}
-
-		if (krb_log_exec(krb5_parse_name_flags, credentials->ctx, sname ? sname : "",
-		                 KRB5_PRINCIPAL_PARSE_NO_REALM, &principal))
-			goto cleanup;
-
-		if (realm)
-		{
-			if (krb_log_exec(krb5glue_set_principal_realm, credentials->ctx, principal, realm))
-				goto cleanup;
-		}
-
-		if (krb_log_exec(krb5_kt_start_seq_get, credentials->ctx, credentials->keytab, &cur))
-			goto cleanup;
-
-		do
-		{
-			krb5_error_code rv = krb_log_exec(krb5_kt_next_entry, credentials->ctx,
-			                                  credentials->keytab, &entry, &cur);
-			if (rv == KRB5_KT_END)
-				break;
-			if (rv != 0)
-				goto cleanup;
-
-			if ((!sname || krb_log_exec_bool(krb5_principal_compare_any_realm, credentials->ctx,
-			                                 principal, entry.principal)) &&
-			    (!realm || krb_log_exec_bool(krb5_realm_compare, credentials->ctx, principal,
-			                                 entry.principal)))
-				break;
-			const krb5_error_code res =
-			    krb_log_exec(krb5glue_free_keytab_entry_contents, credentials->ctx, &entry);
-			memset(&entry, 0, sizeof(entry));
-			if (res != 0)
-				goto cleanup;
-		} while (1);
-
-		if (krb_log_exec(krb5_kt_end_seq_get, credentials->ctx, credentials->keytab, &cur))
-			goto cleanup;
-
-		if (!entry.principal)
-			goto cleanup;
-
-		/* Get the TGT */
-		if (krb_log_exec(krb5_get_init_creds_keytab, credentials->ctx, &creds, entry.principal,
-		                 credentials->keytab, 0, NULL, NULL))
+		if (!retrieveSomeTgt(credentials, target, &creds))
 			goto cleanup;
 
 		if (!kerberos_mk_tgt_token(output_buffer, KRB_TGT_REP, NULL, NULL, &creds.ticket))
@@ -1518,7 +1630,8 @@ static BOOL copy_krb5_data(krb5_data* data, PUCHAR* ptr, ULONG* psize)
 }
 #endif
 
-static SECURITY_STATUS SEC_ENTRY kerberos_DeleteSecurityContext(PCtxtHandle phContext)
+static SECURITY_STATUS
+    SEC_ENTRY kerberos_DeleteSecurityContext(WINPR_ATTR_UNUSED PCtxtHandle phContext)
 {
 #ifdef WITH_KRB5
 	KRB_CONTEXT* context = get_context(phContext);
@@ -1602,6 +1715,108 @@ static SECURITY_STATUS kerberos_ATTR_SIZES(KRB_CONTEXT* context, KRB_CREDENTIALS
 	return SEC_E_OK;
 }
 
+static SECURITY_STATUS kerberos_ATTR_AUTH_IDENTITY(KRB_CONTEXT* context,
+                                                   KRB_CREDENTIALS* credentials,
+                                                   SecPkgContext_AuthIdentity* AuthIdentity)
+{
+	const SecPkgContext_AuthIdentity empty = { 0 };
+
+	WINPR_ASSERT(context);
+	WINPR_ASSERT(context->auth_ctx);
+	WINPR_ASSERT(credentials);
+
+	WINPR_ASSERT(AuthIdentity);
+	*AuthIdentity = empty;
+
+	krb5glue_authenticator authenticator = NULL;
+	krb5_error_code rv = krb_log_exec(krb5_auth_con_getauthenticator, credentials->ctx,
+	                                  context->auth_ctx, &authenticator);
+	if (rv)
+		goto fail;
+
+	{
+		rv = -1;
+
+#if defined(WITH_KRB5_HEIMDAL)
+		const Realm data = authenticator->crealm;
+		if (!data)
+			goto fail;
+		const size_t data_len = length_Realm(&data);
+#else
+		krb5_data* realm_data = krb5_princ_realm(credentials->ctx, authenticator->client);
+		if (!realm_data)
+			goto fail;
+		const char* data = realm_data->data;
+		if (!data)
+			goto fail;
+		const size_t data_len = realm_data->length;
+#endif
+
+		if (data_len > (sizeof(AuthIdentity->Domain) - 1))
+			goto fail;
+		strncpy(AuthIdentity->Domain, data, data_len);
+	}
+
+	{
+#if defined(WITH_KRB5_HEIMDAL)
+		const PrincipalName* principal = &authenticator->cname;
+		const size_t name_length = length_PrincipalName(principal);
+		if (!principal->name_string.val)
+			goto fail;
+		const char* name = *principal->name_string.val;
+#else
+		char* name = NULL;
+		rv = krb_log_exec(krb5_unparse_name_flags, credentials->ctx, authenticator->client,
+		                  KRB5_PRINCIPAL_UNPARSE_NO_REALM, &name);
+		if (rv)
+			goto fail;
+
+		const size_t name_length = strlen(name);
+#endif
+
+		const bool ok = (name_length <= (sizeof(AuthIdentity->User) - 1));
+		if (ok)
+			strncpy(AuthIdentity->User, name, name_length);
+
+		rv = ok ? 0 : -1;
+
+#if !defined(WITH_KRB5_HEIMDAL)
+		krb5_free_unparsed_name(credentials->ctx, name);
+#endif
+	}
+
+fail:
+	krb5glue_free_authenticator(credentials->ctx, authenticator);
+	return krb5_error_to_SECURITY_STATUS(rv);
+}
+
+static SECURITY_STATUS kerberos_ATTR_PACKAGE_INFO(WINPR_ATTR_UNUSED KRB_CONTEXT* context,
+                                                  WINPR_ATTR_UNUSED KRB_CREDENTIALS* credentials,
+                                                  SecPkgContext_PackageInfo* PackageInfo)
+{
+	size_t size = sizeof(SecPkgInfoA);
+	SecPkgInfoA* pPackageInfo =
+	    (SecPkgInfoA*)sspi_ContextBufferAlloc(QuerySecurityPackageInfoIndex, size);
+
+	if (!pPackageInfo)
+		return SEC_E_INSUFFICIENT_MEMORY;
+
+	pPackageInfo->fCapabilities = KERBEROS_SecPkgInfoA.fCapabilities;
+	pPackageInfo->wVersion = KERBEROS_SecPkgInfoA.wVersion;
+	pPackageInfo->wRPCID = KERBEROS_SecPkgInfoA.wRPCID;
+	pPackageInfo->cbMaxToken = KERBEROS_SecPkgInfoA.cbMaxToken;
+	pPackageInfo->Name = _strdup(KERBEROS_SecPkgInfoA.Name);
+	pPackageInfo->Comment = _strdup(KERBEROS_SecPkgInfoA.Comment);
+
+	if (!pPackageInfo->Name || !pPackageInfo->Comment)
+	{
+		sspi_ContextBufferFree(pPackageInfo);
+		return SEC_E_INSUFFICIENT_MEMORY;
+	}
+	PackageInfo->PackageInfo = pPackageInfo;
+	return SEC_E_OK;
+}
+
 static SECURITY_STATUS kerberos_ATTR_TICKET_LOGON(KRB_CONTEXT* context,
                                                   KRB_CREDENTIALS* credentials,
                                                   KERB_TICKET_LOGON* ticketLogon)
@@ -1646,27 +1861,29 @@ again:
 	if (krb_log_exec(krb5_auth_con_init, credentials->ctx, &authContext))
 		goto out;
 
-	krb5_data derOut = { 0 };
-	if (krb_log_exec(krb5_fwd_tgt_creds, credentials->ctx, authContext, context->targetHost,
-	                 matchCred.client, matchCred.server, credentials->ccache, 1, &derOut))
 	{
-		ret = SEC_E_LOGON_DENIED;
-		goto out;
+		krb5_data derOut = { 0 };
+		if (krb_log_exec(krb5_fwd_tgt_creds, credentials->ctx, authContext, context->targetHost,
+		                 matchCred.client, matchCred.server, credentials->ccache, 1, &derOut))
+		{
+			ret = SEC_E_LOGON_DENIED;
+			goto out;
+		}
+
+		ticketLogon->MessageType = KerbTicketLogon;
+		ticketLogon->Flags = KERB_LOGON_FLAG_REDIRECTED;
+
+		if (!copy_krb5_data(&hostCred->ticket, &ticketLogon->ServiceTicket,
+		                    &ticketLogon->ServiceTicketLength))
+		{
+			krb5_free_data(credentials->ctx, &derOut);
+			goto out;
+		}
+
+		ticketLogon->TicketGrantingTicketLength = derOut.length;
+		ticketLogon->TicketGrantingTicket = (PUCHAR)derOut.data;
 	}
 
-	ticketLogon->MessageType = KerbTicketLogon;
-	ticketLogon->Flags = KERB_LOGON_FLAG_REDIRECTED;
-
-	if (!copy_krb5_data(&hostCred->ticket, &ticketLogon->ServiceTicket,
-	                    &ticketLogon->ServiceTicketLength))
-	{
-		krb5_free_data(credentials->ctx, &derOut);
-		goto out;
-	}
-
-	ticketLogon->TicketGrantingTicketLength = derOut.length;
-	ticketLogon->TicketGrantingTicket = (PUCHAR)derOut.data;
-
 	ret = SEC_E_OK;
 out:
 	krb5_auth_con_free(credentials->ctx, authContext);
@@ -1677,8 +1894,9 @@ out:
 
 #endif /* WITH_KRB5 */
 
-static SECURITY_STATUS SEC_ENTRY kerberos_QueryContextAttributesA(PCtxtHandle phContext,
-                                                                  ULONG ulAttribute, void* pBuffer)
+static SECURITY_STATUS
+    SEC_ENTRY kerberos_QueryContextAttributesA(PCtxtHandle phContext,
+                                               WINPR_ATTR_UNUSED ULONG ulAttribute, void* pBuffer)
 {
 	if (!phContext)
 		return SEC_E_INVALID_HANDLE;
@@ -1698,6 +1916,14 @@ static SECURITY_STATUS SEC_ENTRY kerberos_QueryContextAttributesA(PCtxtHandle ph
 		case SECPKG_ATTR_SIZES:
 			return kerberos_ATTR_SIZES(context, credentials, (SecPkgContext_Sizes*)pBuffer);
 
+		case SECPKG_ATTR_AUTH_IDENTITY:
+			return kerberos_ATTR_AUTH_IDENTITY(context, credentials,
+			                                   (SecPkgContext_AuthIdentity*)pBuffer);
+
+		case SECPKG_ATTR_PACKAGE_INFO:
+			return kerberos_ATTR_PACKAGE_INFO(context, credentials,
+			                                  (SecPkgContext_PackageInfo*)pBuffer);
+
 		case SECPKG_CRED_ATTR_TICKET_LOGON:
 			return kerberos_ATTR_TICKET_LOGON(context, credentials, (KERB_TICKET_LOGON*)pBuffer);
 
@@ -1731,10 +1957,10 @@ static SECURITY_STATUS SEC_ENTRY kerberos_SetContextAttributesA(
 	return SEC_E_UNSUPPORTED_FUNCTION;
 }
 
-static SECURITY_STATUS SEC_ENTRY kerberos_SetCredentialsAttributesX(PCredHandle phCredential,
-                                                                    ULONG ulAttribute,
-                                                                    void* pBuffer, ULONG cbBuffer,
-                                                                    WINPR_ATTR_UNUSED BOOL unicode)
+static SECURITY_STATUS SEC_ENTRY kerberos_SetCredentialsAttributesX(
+    WINPR_ATTR_UNUSED PCredHandle phCredential, WINPR_ATTR_UNUSED ULONG ulAttribute,
+    WINPR_ATTR_UNUSED void* pBuffer, WINPR_ATTR_UNUSED ULONG cbBuffer,
+    WINPR_ATTR_UNUSED BOOL unicode)
 {
 #ifdef WITH_KRB5
 	KRB_CREDENTIALS* credentials = NULL;
@@ -1809,9 +2035,10 @@ static SECURITY_STATUS SEC_ENTRY kerberos_SetCredentialsAttributesA(PCredHandle
 	return kerberos_SetCredentialsAttributesX(phCredential, ulAttribute, pBuffer, cbBuffer, FALSE);
 }
 
-static SECURITY_STATUS SEC_ENTRY kerberos_EncryptMessage(PCtxtHandle phContext, ULONG fQOP,
-                                                         PSecBufferDesc pMessage,
-                                                         ULONG MessageSeqNo)
+static SECURITY_STATUS SEC_ENTRY kerberos_EncryptMessage(WINPR_ATTR_UNUSED PCtxtHandle phContext,
+                                                         WINPR_ATTR_UNUSED ULONG fQOP,
+                                                         WINPR_ATTR_UNUSED PSecBufferDesc pMessage,
+                                                         WINPR_ATTR_UNUSED ULONG MessageSeqNo)
 {
 #ifdef WITH_KRB5
 	KRB_CONTEXT* context = get_context(phContext);
@@ -1899,9 +2126,10 @@ static SECURITY_STATUS SEC_ENTRY kerberos_EncryptMessage(PCtxtHandle phContext,
 #endif
 }
 
-static SECURITY_STATUS SEC_ENTRY kerberos_DecryptMessage(PCtxtHandle phContext,
-                                                         PSecBufferDesc pMessage,
-                                                         ULONG MessageSeqNo, ULONG* pfQOP)
+static SECURITY_STATUS SEC_ENTRY kerberos_DecryptMessage(WINPR_ATTR_UNUSED PCtxtHandle phContext,
+                                                         WINPR_ATTR_UNUSED PSecBufferDesc pMessage,
+                                                         WINPR_ATTR_UNUSED ULONG MessageSeqNo,
+                                                         WINPR_ATTR_UNUSED ULONG* pfQOP)
 {
 #ifdef WITH_KRB5
 	KRB_CONTEXT* context = get_context(phContext);
@@ -2005,9 +2233,10 @@ static SECURITY_STATUS SEC_ENTRY kerberos_DecryptMessage(PCtxtHandle phContext,
 #endif
 }
 
-static SECURITY_STATUS SEC_ENTRY kerberos_MakeSignature(PCtxtHandle phContext,
+static SECURITY_STATUS SEC_ENTRY kerberos_MakeSignature(WINPR_ATTR_UNUSED PCtxtHandle phContext,
                                                         WINPR_ATTR_UNUSED ULONG fQOP,
-                                                        PSecBufferDesc pMessage, ULONG MessageSeqNo)
+                                                        WINPR_ATTR_UNUSED PSecBufferDesc pMessage,
+                                                        WINPR_ATTR_UNUSED ULONG MessageSeqNo)
 {
 #ifdef WITH_KRB5
 	KRB_CONTEXT* context = get_context(phContext);
@@ -2076,9 +2305,9 @@ static SECURITY_STATUS SEC_ENTRY kerberos_MakeSignature(PCtxtHandle phContext,
 #endif
 }
 
-static SECURITY_STATUS SEC_ENTRY kerberos_VerifySignature(PCtxtHandle phContext,
-                                                          PSecBufferDesc pMessage,
-                                                          ULONG MessageSeqNo,
+static SECURITY_STATUS SEC_ENTRY kerberos_VerifySignature(WINPR_ATTR_UNUSED PCtxtHandle phContext,
+                                                          WINPR_ATTR_UNUSED PSecBufferDesc pMessage,
+                                                          WINPR_ATTR_UNUSED ULONG MessageSeqNo,
                                                           WINPR_ATTR_UNUSED ULONG* pfQOP)
 {
 #ifdef WITH_KRB5
